__all__ = [
    "consts",
    "db",
    "grid",
    "io",
    "radex",
    "LAMDA",
    "run",
    "save_dataset",
    "load_dataset",
]
__version__ = "0.3.1"
__author__ = "Akio Taniguchi"


# submodules
from . import consts
from . import db
from . import grid
from . import io
from . import radex
from .db import *
from .grid import *
from .io import *
from .radex import *


# builtin RADEX binaries
radex.build()



__all__ = [
    "NDRADEX",
    "NDRADEX_CONFIG",
    "RADEX_BIN",
    "RADEX_VERSION",
    "LAMDA_ALIASES",
    "DV",
    "GEOM",
    "N_H2",
    "N_MOL",
    "T_BG",
    "T_KIN",
    "TIMEOUT",
    "N_E",
    "N_H",
    "N_HE",
    "N_HP",
    "N_OH2",
    "N_PH2",
]


# standard library
from os import getenv
from pathlib import Path
from typing import Dict, TypeVar


# dependencies
from tomlkit import load


# type hints
T = TypeVar("T")


# constants
NDRADEX = Path(__file__).parent
"""Path of the ndRADEX package"""


# helper functions
def ensure(toml: Path) -> Path:
    """Create an empty TOML file if it does not exist."""
    if not toml.exists():
        toml.parent.mkdir(parents=True, exist_ok=True)
        toml.touch()

    return toml


def getval(toml: Path, keys: str, default: T) -> T:
    """Return the value of the keys in a TOML file."""
    with open(toml, "r") as file:
        doc = load(file)

    for key in keys.split("."):
        if (doc := doc.get(key)) is None:
            return default

    return type(default)(doc.unwrap())


# config file and directory for it
if (env := getenv("NDRADEX_DIR")) is not None:
    NDRADEX_DIR = Path(env)
elif (env := getenv("XDG_CONFIG_HOME")) is not None:
    NDRADEX_DIR = Path(env) / "ndradex"
else:
    NDRADEX_DIR = Path.home() / ".config" / "ndradex"

NDRADEX_CONFIG = ensure(NDRADEX_DIR / "config.toml")


# query aliases for molecular/atomic databases
LAMDA_ALIASES: Dict[str, str] = getval(NDRADEX_CONFIG, "lamda.aliases", {})
"""Query aliases for the LAMDA database."""


# default values for RADEX binaries
RADEX_BIN = getval(NDRADEX_CONFIG, "defaults.bin", NDRADEX / "bin")
"""Default path for the RADEX binaries."""

RADEX_VERSION = "30nov2011"
"""Supported version of the RADEX binaries."""


# default values for public functions
DV = getval(NDRADEX_CONFIG, "defaults.dv", 1.0)
"""Default value for the ``dv`` argument."""

GEOM = getval(NDRADEX_CONFIG, "defaults.geom", "uni")
"""Default value for the ``geom`` argument."""

N_H2 = getval(NDRADEX_CONFIG, "defaults.n_H2", 1e3)
"""Default value for the ``n_H2`` argument."""

N_MOL = getval(NDRADEX_CONFIG, "defaults.N_mol", 1e15)
"""Default value for the ``N_mol`` argument."""

N_PARALLEL = getval(NDRADEX_CONFIG, "defaults.n_parallel", 2)
"""Default value for the ``n_parallel`` argument."""

PROGRESS = getval(NDRADEX_CONFIG, "defaults.progress", True)
"""Default value for the ``progress`` argument."""

SQUEEZE = getval(NDRADEX_CONFIG, "defaults.squeeze", True)
"""Default value for the ``squeeze`` argument."""

TIMEOUT = getval(NDRADEX_CONFIG, "defaults.timeout", 30.0)
"""Default value for the ``timeout`` argument."""

T_BG = getval(NDRADEX_CONFIG, "defaults.T_bg", 2.73)
"""Default value for the ``T_bg`` argument."""

T_KIN = getval(NDRADEX_CONFIG, "defaults.T_kin", 100.0)
"""Default value for the ``T_kin`` argument."""


# default values for public functions (optional)
N_E = getval(NDRADEX_CONFIG, "defaults.n_e", 0.0) or None
"""Default value for the ``n_e`` argument."""

N_H = getval(NDRADEX_CONFIG, "defaults.n_H", 0.0) or None
"""Default value for the ``n_H`` argument."""

N_HE = getval(NDRADEX_CONFIG, "defaults.n_He", 0.0) or None
"""Default value for the ``n_He`` argument."""

N_HP = getval(NDRADEX_CONFIG, "defaults.n_Hp", 0.0) or None
"""Default value for the ``n_Hp`` argument."""

N_OH2 = getval(NDRADEX_CONFIG, "defaults.n_oH2", 0.0) or None
"""Default value for the ``n_oH2`` argument."""

N_PH2 = getval(NDRADEX_CONFIG, "defaults.n_pH2", 0.0) or None
"""Default value for the ``n_pH2`` argument."""



__all__ = ["LAMDA"]


# tandard library
import re
import warnings
from functools import partial, wraps
from logging import getLogger
from pathlib import Path
from urllib.parse import urlparse


# dependencies
import numpy as np
from .consts import LAMDA_ALIASES


logger = getLogger(__name__)


def cache(func):
    """Decorator for caching result of a method."""
    name = "_" + func.__name__

    @wraps(func)
    def wrapped(self):
        if hasattr(self, name):
            return getattr(self, name)

        result = func(self)
        setattr(self, name, result)
        return result

    return wrapped


class LAMDA:
    def __init__(self, query, dir="."):
        query = LAMDA_ALIASES.get(query, query)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            tables = get_tables(query)
            path = get_data_path(query, dir)

        self._collrates = tables[0]
        self._transitions = tables[1]
        self._levels = tables[2]
        self._data_path = path

        self.desc = self._levels.meta["molecule"]

    @property
    @cache
    def qn_ul(self):
        """List of transition quantum numbers."""
        return list(self._transitions["QN_ul"])

    @property
    @cache
    def freq(self):
        """Transition frequencies in units of GHz."""
        freq = self._transitions["Frequency"]
        return dict(zip(self.qn_ul, freq))

    @property
    @cache
    def freq_lim(self):
        """Transition frequency ranges in units of GHz."""
        freq = self._transitions["Frequency"]
        freq_lim = [f"{(1-1e-9)*f} {(1+1e-9)*f}" for f in freq]
        return dict(zip(self.qn_ul, freq_lim))

    @property
    @cache
    def a_coeff(self):
        """Einstein A coefficients in units of s^-1."""
        a_coeff = self._transitions["EinsteinA"]
        return dict(zip(self.qn_ul, a_coeff))

    @property
    @cache
    def e_up(self):
        """Upper state energies in units of K."""
        e_up = self._transitions["E_u(K)"]
        return dict(zip(self.qn_ul, e_up))

    @property
    @cache
    def n_crit(self):
        """Critical densities in units of cm^-3."""
        # lazy import of astropy-related things
        from astroquery.lamda.utils import ncrit

        @np.vectorize
        def func(T_kin, index_u, index_l):
            tables = (self._collrates, self._transitions, self._levels)
            return ncrit(tables, index_u, index_l, T_kin).value

        funcs = {}

        for qn_ul in self.qn_ul:
            index_u = self._transitions.loc[qn_ul]["Upper"]
            index_l = self._transitions.loc[qn_ul]["Lower"]

            funcs[qn_ul] = partial(func, index_u=index_u, index_l=index_l)

        return funcs

    def __enter__(self):
        """Create a temporary LAMDA data inside a context block."""
        # lazy import of astropy-related things
        from astroquery.lamda import write_lamda_datafile

        tables = (self._collrates, self._transitions, self._levels)
        write_lamda_datafile(self._data_path, tables)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Delete a temporary LAMDA data outside a context block."""
        self._data_path.unlink()

    def __str__(self):
        return str(self._data_path)

    def __repr__(self):
        return f"LAMDA({self.desc})"


# utility functions
def get_tables(query):
    """(Down)load LAMDA data as astropy tables.

    This will also add a column of transition quantum
    numbers (i.e., 1-0) to the transition table (QN_ul).

    """
    # lazy import of astropy-related things
    from astropy.table import Column

    collrates, transitions, levels = get_raw_tables(query)
    levels.add_index("Level")

    data = []
    for row in transitions:
        qn_u = format_qn(levels.loc[row["Upper"]]["J"])
        qn_l = format_qn(levels.loc[row["Lower"]]["J"])
        data.append(f"{qn_u}-{qn_l}")

    transitions.add_column(Column(data, "QN_ul"))
    transitions.add_index("QN_ul")
    return collrates, transitions, levels


def get_raw_tables(query):
    """(Down)load LAMDA data as astropy tables."""
    # lazy import of astropy-related things
    from astroquery.lamda import Lamda
    from astroquery.lamda import parse_lamda_datafile

    # case 1: try to get by URL
    if query.startswith("http"):
        name = Path(urlparse(query).path).stem
        Lamda.molecule_dict[name] = query
        return Lamda.query(name)

    # case 2: try to get by local path
    path = Path(query).expanduser().resolve()
    if path.exists():
        return parse_lamda_datafile(path)

    # case 3: try to get by astroquery
    try:
        return Lamda.query(Path(query).stem)
    except Exception:
        raise ValueError(query)


def get_data_path(query, dir="."):
    """Get path object for temporary LAMDA data."""
    data = Path(query).stem + ".dat"
    return Path(dir, data).expanduser().resolve()


def format_qn(qn):
    """Format QN string to be well structured."""
    # trim single/double quotes
    qn = re.sub(r'\'|"', "", qn)

    # trim space of both ends
    qn = re.sub(r"^\s+|\s+$", "", qn)

    # convert space or underscore to comma
    qn = re.sub(r"\s+|_+", ",", qn)

    # trim unnecessary zero of int (e.g., 01 -> 1)
    pat = re.compile(r"^0([0-9])$")
    qn = ",".join(pat.sub(r"\1", _) for _ in qn.split(","))

    # trim unnecessary zero of float (e.g., 3.50 -> 3.5)
    pat = re.compile(r"([0-9]+.[0-9]+)0+")
    qn = ",".join(pat.sub(r"\1", _) for _ in qn.split(","))

    # add parenthesis if at least one comma exists
    return re.sub(r"(.*,.*)", r"(\1)", qn)


def list_available(path, max_transitions=None):
    """List names of datafiles and transitions in Markdown."""
    # lazy import of astropy-related things
    from astroquery.lamda import Lamda

    def sorter(item):
        name = re.sub(r"[a-z]-(.*)", r"\1", item[0])
        return "".join(re.findall(r"[a-z]", name))

    names, descs, trans = [], [], []
    for name, url in sorted(Lamda.molecule_dict.items(), key=sorter):
        try:
            lamda = LAMDA(name)
        except Exception:
            continue

        names.append(f"[`{name}.dat`]({url})")
        descs.append(lamda.desc)
        trans.append([f"`{q}`" for q in lamda.qn_ul])

    with open(path, "w") as f:
        f.write("| Query name | Description | Transitions (QN_ul) |\n")
        f.write("| --- | --- | --- |\n")

        for name, desc, tran in zip(names, descs, trans):
            if max_transitions is None or len(tran) <= max_transitions:
                tran = ", ".join(tran)
            else:
                tran = ", ".join(tran[:max_transitions]) + ", ..."

            f.write(f"| {name} | {desc} | {tran} |\n")



__all__ = []


# standard library
from logging import getLogger
from pathlib import Path
from subprocess import PIPE, run as sprun
from subprocess import CalledProcessError, TimeoutExpired
from typing import List, Optional, Sequence, Union


# dependencies
from .consts import NDRADEX, RADEX_VERSION


# type hints
PathLike = Union[Path, str]


# constants
N_OUTPUT_VALUES = 10


# module logger
logger = getLogger(__name__)


def run(
    input: Sequence[str],
    radex: PathLike,
    cleanup: bool = True,
    timeout: Optional[float] = None,
) -> List[str]:
    """Run RADEX and return the output as a list of strings.

    Note that the function only reads the last line of a RADEX output.
    This means that the function will only return the values
    of the highest-frequency transition specified in the RADEX input.

    Args:
        input: Input strings for RADEX. See the examples below.
        radex: Absolute path of the RADEX binary to be run.
        cleanup: If True, the RADEX output file will be deleted.
        timeout: Timeout of the run in units of seconds.
            Default is None (unlimited run time is allowed).

    Returns:
        RADEX output as a list of strings.

    Examples:
        To get output of CO(1-0) @ T_kin = 100 K, n_H2 = 1e3 cm^-3,
        N_CO = 1e15 cm^-2, T_bg = 2.73 K, and dv = 1.0 km s^-1::

            input = [
                "co.dat", "radex.out", "110 120", "100",
                "1", "H2", "1e3", "2.73", "1e15", "1.0", "0",
            ]
            output = run(input, "/path/to/radex")

    """
    try:
        sprun(
            str(radex),
            input="\n".join(input),
            timeout=timeout,
            stdout=PIPE,
            stderr=PIPE,
            check=True,
            text=True,
        )
        return finalize(input[1], cleanup)
    except (IndexError, TypeError):
        logger.warning("RADEX did not run due to invalid input")
        return ["NaN"] * N_OUTPUT_VALUES
    except FileNotFoundError:
        logger.warning("RADEX path or moldata does not exist")
        return ["NaN"] * N_OUTPUT_VALUES
    except CalledProcessError:
        logger.warning("RADEX failed due to invalid input")
        return ["NaN"] * N_OUTPUT_VALUES
    except TimeoutExpired:
        logger.warning("RADEX interrupted due to timeout")
        return ["NaN"] * N_OUTPUT_VALUES
    except RuntimeError:
        logger.warning("RADEX version is not valid")
        return ["NaN"] * N_OUTPUT_VALUES


def build(force: bool = False) -> None:
    """Build the builtin RADEX binaries."""
    if force:
        sprun(
            args=["make", "clean"],
            cwd=NDRADEX / "bin",
            stdout=PIPE,
            stderr=PIPE,
            check=True,
        )

    sprun(
        args=[
            "make",
            "build",
            "RADEX_LOGFILE=/dev/null",
            "RADEX_MAXITER=999999",
        ],
        cwd=NDRADEX / "bin",
        stdout=PIPE,
        stderr=PIPE,
        check=True,
    )


def finalize(output: PathLike, cleanup: bool) -> List[str]:
    """Read a RADEX output file and return the final output."""
    try:
        with open(output) as f:
            lines = f.readlines()

        if RADEX_VERSION not in lines[0]:
            raise RuntimeError("RADEX version is not valid")

        return lines[-1].split()[-N_OUTPUT_VALUES:]
    finally:
        if cleanup:
            Path(output).unlink(missing_ok=True)



__all__ = ["run"]


# standard library
from concurrent.futures import ProcessPoolExecutor
from enum import Enum, auto
from itertools import product, repeat
from pathlib import Path
from random import getrandbits
from tempfile import TemporaryDirectory


# dependencies
import numpy as np
import pandas as pd
import xarray as xr
from tqdm import tqdm
from .consts import (
    DV,
    GEOM,
    N_E,
    N_H,
    N_H2,
    N_HE,
    N_HP,
    N_MOL,
    N_OH2,
    N_PARALLEL,
    N_PH2,
    PROGRESS,
    RADEX_BIN,
    SQUEEZE,
    T_BG,
    T_KIN,
    TIMEOUT,
)
from .db import LAMDA
from .radex import run as run_radex


# constants
class Dims(Enum):
    QN_ul = "QN_ul"
    T_kin = "T_kin"
    N_mol = "N_mol"
    n_H2 = "H2"
    n_pH2 = "p-H2"
    n_oH2 = "o-H2"
    n_e = "e"
    n_H = "H"
    n_He = "He"
    n_Hp = "H+"
    T_bg = "T_bg"
    dv = "dv"
    geom = "geom"


class Vars(Enum):
    E_u = auto()
    freq = auto()
    wavel = auto()
    T_ex = auto()
    tau = auto()
    T_r = auto()
    pop_u = auto()
    pop_l = auto()
    I = auto()
    F = auto()


# main function
def run(
    query,
    QN_ul,
    T_kin=T_KIN,
    N_mol=N_MOL,
    n_H2=N_H2,
    n_pH2=N_PH2,
    n_oH2=N_OH2,
    n_e=N_E,
    n_H=N_H,
    n_He=N_HE,
    n_Hp=N_HP,
    T_bg=T_BG,
    dv=DV,
    geom=GEOM,
    *,
    squeeze=SQUEEZE,
    progress=PROGRESS,
    timeout=TIMEOUT,
    n_procs=N_PARALLEL,
    work_dir=None,
):
    """Run grid RADEX calculation and get results as xarray.Dataset.

    This is the main function of ndRADEX. It provides 13 parameters
    which can be griddable (i.e., both scalar and array are accepted).
    The output is an xarray's Dataset of multi-dimensional DataArrays
    whose shapes are the product of gridded parameters (i.e., argument
    of parameter is array-like and its length is more than one).
    For example, if you run grid RADEX calculation for CO (1-0) with
    kinetic temperatures of [100, 200, 300, 400, 500] K, and H2
    densities of [1e3, 1e4, 1e5] cm^-3, then the shepe is (5, 3).
    For more information: https://github.com/astropenguin/ndradex/wiki

    Args:
        query (str): Name of LAMDA datafile (e.g., 'co' or 'co.dat').
            You can also specify the path or URL of the datafile.
        QN_ul (str, griddable): Name(s) of transition (e.g., '1-0').
        T_kin (int or float, griddable): Value(s) of kinetic temperature
            in units of K.
        N_mol (int or float, griddable): Value(s) of column density of
            molecule (or atom) in units of cm^-2.
        n_H2 (int or float, griddable): Value(s) of H2 density in units
            of cm^-3. Do not specify together with n_pH2 and n_OH2.
        n_pH2 (int or float, griddable): Value(s) of para-H2 density
            in units of cm^-3. Do not specify together with n_H2.
            This parameter is not activated by default.
        n_oH2 (int or float, griddable): Value(s) of ortho-H2 density
            in units of cm^-3. Do not specify together with n_H2.
            This parameter is not activated by default.
        n_e (int or float, griddable): Value(s) of electron density
            in units of cm^-3. This parameter is not activated by default.
        n_H (int or float, griddable): Value(s) of atomic hydrogen density
            in units of cm^-3. This parameter is not activated by default.
        n_He (int or float, griddable): Value(s) of atomic helium density
            in units of cm^-3. This parameter is not activated by default.
        n_Hp (int or float, griddable): Value(s) of ionized hydrogen density
            in units of cm^-3. This parameter is not activated by default.
        T_bg (int or float, griddable): Value(s) of background temperature
            in units of K.
        dv (int or float, griddable): Value(s) of FWHM width of a line
            in units of km s^-1.
        geom (str, griddable): Name(s) of geometry for photon escape
            probability. Either 'uni', 'lvg', or 'slab' is acccepted.
        squeeze (bool, optional): If True (default), then dimensions
            whose length is only one are dropped.
        progress (bool, optional): If True (default), then a bar is
            shown during the calculation to show the progress.
        timeout (int, optional): Timeout of a RADEX run in units of second.
            Default is None (unlimited run time is permitted).
        n_procs (int, optional): Number of processes for asynchronous
            RADEX calculations. Default is None (<number of CPU count>-1).
        work_dir (path-like, optional): Path of working directory under
            which a temporary directory for intermediate files is created.
            The temporary directory is sure to be automatically created and
            deleted even if the function is interrupted by some errors.
            Default is None (the OS-native temporary directory is used).

    Returns:
        dataset (xarray.Dataset): Dataset which contains DataArrays of:
            E_u: upper state energy in units of K,
            freq: transition frequency in units of GHz,
            wavel: transition wavelength in units of um,
            T_ex: excitation temperature in units of K,
            tau: opacity of line,
            T_r: peak intensity in units of K
            pop_u: upper state population
            pop_l: lower state population,
            I: flux of line in units of K km s^-1,
            F: flux of line in units of erg s^-1 cm^-2.

    Examples:
        To get the values of CO J=[1-0, 2-1] @ T_kin = [100, 200, 300,
        400, 500] K, n_H2 = [1e3, 1e4, 1e5] cm^-3, N_CO = 1e15 cm^-2,
        T_bg = 2.73 K, and dv = 1.0 km s^-1:

            >>> dataset = run('co', ['1-0', '2-1'],
                              T_kin=[100, 200, 300, 400, 500],
                              n_H2=[1e3, 1e4, 1e5],
                              N_mol=1e15, T_bg=2.73, dv=1.0)
            >>> print(dataset)
            <xarray.Dataset>
            Dimensions:      (QN_ul: 2, T_kin: 5, n_H2: 3)
            Coordinates:
            * QN_ul        (QN_ul) <U3 '1-0' '2-1'
            * T_kin        (T_kin) int64 100 200 300 400 500
                N_mol        float64 1e+15
            * n_H2         (n_H2) float64 1e+03 1e+04 1e+05
                T_bg         float64 2.73
                dv           float64 1.0
                geom         <U3 'uni'
                description  <U9 'LAMDA(CO)'
            Data variables:
                E_u          (QN_ul, T_kin, n_H2) float64 5.5 ...
                freq         (QN_ul, T_kin, n_H2) float64 115.3 ...
                wavel        (QN_ul, T_kin, n_H2) float64 2.601e+03 ...
                T_ex         (QN_ul, T_kin, n_H2) float64 132.5 ...
                tau          (QN_ul, T_kin, n_H2) float64 0.009966 ...
                T_r          (QN_ul, T_kin, n_H2) float64 1.278 ...
                pop_u        (QN_ul, T_kin, n_H2) float64 0.4934 ...
                pop_l        (QN_ul, T_kin, n_H2) float64 0.1715 ...
                I            (QN_ul, T_kin, n_H2) float64 1.36 ...
                F            (QN_ul, T_kin, n_H2) float64 2.684e-08 ...

    """
    empty = get_empty_array(
        QN_ul,
        T_kin,
        N_mol,
        n_H2,
        n_pH2,
        n_oH2,
        n_e,
        n_H,
        n_He,
        n_Hp,
        T_bg,
        dv,
        geom,
    )

    with TemporaryDirectory(dir=work_dir) as temp_dir:
        with LAMDA(query, temp_dir) as lamda:
            # make an empty dataset and flattened args
            dataset = get_empty_dataset(lamda, empty)
            iterables = [
                generate_inputs(lamda, empty),
                generate_radex_paths(lamda, empty),
                repeat(timeout),
            ]

            # run RADEX with multiprocess and update dataset
            execute(
                dataset,
                *iterables,
                dir=temp_dir,
                progress=progress,
                n_procs=n_procs,
            )

    return finalize(dataset, squeeze)


# sub functions
def generate_inputs(lamda, empty):
    """Generate RADEX input string iteratively."""
    template = get_input_template(lamda, empty)

    for kwargs in generate_kwargs(lamda, empty):
        yield template.format(**kwargs).split("\n")


def generate_radex_paths(lamda, empty):
    """Generate RADEX path iteratively."""
    path = str(RADEX_BIN / "radex-{geom}")

    for kwargs in generate_kwargs(lamda, empty):
        yield path.format(geom=kwargs[Dims.geom.name])


def get_empty_dataset(lamda, empty):
    """Make an empty xarray.Dataset for storing DataArrays."""
    dataset = xr.Dataset({var.name: empty.copy() for var in Vars})
    dataset.coords["description"] = repr(lamda)
    return dataset


def execute(dataset, *iterables, dir=".", progress=True, n_procs=None):
    """Run grid RADEX calculation and store results into a dataset."""
    total = np.prod(list(dataset.dims.values()))
    outfile = Path(dir, "grid.out").expanduser().resolve()

    # fmt: off
    with tqdm(total=total, disable=not progress) as bar, \
         ProcessPoolExecutor(n_procs) as executor, \
         outfile.open("w", buffering=1) as f:
    # fmt: on
        for output in executor.map(run_radex, *iterables):
            f.write(",".join(output) + "\n")
            bar.update(1)

    names = [var.name for var in Vars]
    df = pd.read_csv(outfile, header=None, names=names)

    for name in names:
        da = dataset[name]
        da.values = df[name].to_numpy().reshape(da.shape)


def finalize(dataset, squeeze=True):
    """Do finalization before returning a dataset."""
    if not squeeze:
        return dataset

    for dim in dataset.dims:
        coord = dataset.coords[dim]

        if np.issubdtype(coord.dtype, np.str_):
            continue

        if np.any(np.isnan(coord)):
            del dataset.coords[dim]

    return dataset.squeeze()


# utility functions
def get_empty_array(
    QN_ul,
    T_kin,
    N_mol,
    n_H2,
    n_pH2,
    n_oH2,
    n_e,
    n_H,
    n_He,
    n_Hp,
    T_bg,
    dv,
    geom,
):
    """Make an empty xarray.DataArray for storing grid RADEX results."""
    values = {
        Dims.QN_ul: ensure_values(QN_ul),
        Dims.T_kin: ensure_values(T_kin, "K"),
        Dims.N_mol: ensure_values(N_mol, "cm^-2"),
        Dims.n_H2: ensure_values(n_H2, "cm^-3"),
        Dims.n_pH2: ensure_values(n_pH2, "cm^-3"),
        Dims.n_oH2: ensure_values(n_oH2, "cm^-3"),
        Dims.n_e: ensure_values(n_e, "cm^-3"),
        Dims.n_H: ensure_values(n_H, "cm^-3"),
        Dims.n_He: ensure_values(n_He, "cm^-3"),
        Dims.n_Hp: ensure_values(n_Hp, "cm^-3"),
        Dims.T_bg: ensure_values(T_bg, "K"),
        Dims.dv: ensure_values(dv, "km/s"),
        Dims.geom: ensure_values(geom),
    }

    coords = [(dim.name, values[dim]) for dim in Dims]
    shape = [c[1].size for c in coords]

    return xr.DataArray(np.empty(shape), coords)


def generate_kwargs(lamda, empty):
    """Generate keyword args for RADEX input iteratively."""
    freq_lim = lamda.freq_lim

    dims, coords = empty.dims, empty.coords
    flatargs = product(*(coords[dim].values for dim in dims))

    for args in flatargs:
        kwargs = dict(zip(dims, args))
        kwargs["id"] = get_random_hex(8)
        kwargs["freq_lim"] = freq_lim[kwargs["QN_ul"]]
        yield kwargs


def get_input_template(lamda, empty):
    """Make template string for RADEX input."""
    n_dims = [
        dim
        for dim in empty.dims
        if dim.startswith("n_") and not np.any(np.isnan(empty.coords[dim]))
    ]

    template = [
        f"{lamda}",
        f"{lamda}.{{id}}.out",
        "{freq_lim}",
        "{T_kin}",
        str(len(n_dims)),
    ]

    for dim in n_dims:
        template.extend([Dims[dim].value, f"{{{dim}}}"])

    template.extend(["{T_bg}", "{N_mol}", "{dv}", "0"])
    return "\n".join(template)


def ensure_values(values, unit=None):
    """Ensure the type of output and the unit of values."""
    # lazy import of astropy-related things
    from astropy import units as u

    if values is None:
        values = np.nan

    if isinstance(values, u.Quantity):
        unit = u.Unit(unit)
        values = values.to(unit)

    values = np.asarray(values)

    if values.size == 1 and not values.shape:
        return values[np.newaxis]
    else:
        return values


def get_random_hex(length: int = 8) -> str:
    """Return a random hexadecimal string of given length."""
    return f"{getrandbits(length * 4):x}"



__all__ = ["save_dataset", "load_dataset"]


# standard library
from pathlib import Path
from typing import Union


# dependencies
import xarray as xr


# type hints
PathLike = Union[Path, str]


def save_dataset(dataset: xr.Dataset, path: PathLike) -> None:
    """Save an xarray Dataset to a netCDF."""
    dataset.to_netcdf(path)


def load_dataset(path: PathLike) -> xr.Dataset:
    """Load a netCDF to an xarray Dataset."""
    return xr.open_dataset(path)
